Attribute VB_Name = "MGeo"
Option Explicit

'Geometrie-Typen
'fr die Typen braucht man ein eigenes Modul,
'da sonst der Fehler auftritt: "gegenseitige Abhngigkeit von Modulen"
'2 = 2-dimensional; F = Float (Double)

'stellt einen 2D-Punkt dar
'wird oft auch mit Vertex bezeichnet
Public Type Point2F
    X As Double
    Y As Double
End Type

'stellt eine 2D-Strecke dar
Public Type LineSegment2F
    P(0 To 1) As Point2F
End Type

'stellt ein 2D-Viereck dar
Public Type Quad2F
    P(0 To 3) As Point2F
End Type
Public EraseQuad2F As Quad2F

'stellt ein 2D-Polygon dar
Public Type Polygon2F
    Count As Long
    P()   As Point2F
End Type
Public ErasePolygon2F As Polygon2F

Public Enum Position2
    RightPosition = -1
    Collinear = 0
    LeftPosition = 1
End Enum
Private Const eps As Double = 0.00000000000001

Public Function IntersectionPoint(ByVal x1 As Double, ByVal y1 As Double, _
                                  ByVal x2 As Double, ByVal y2 As Double, _
                                  ByVal x3 As Double, ByVal y3 As Double, _
                                  ByVal x4 As Double, ByVal y4 As Double) As Point2F
    'gibt den Schnittpunkt zweier Geraden G1, G2 zurck
    'G1 gegeben durch die Punkte G1:{{x1, y1}, {x2, y2}}
    'G2 gegeben durch die Punkte G2:{{x3, y3}, {x4, y4}}
    Dim Dx1 As Double, Dx2 As Double, Dx3 As Double
    Dim Dy1 As Double, Dy2 As Double, Dy3 As Double
    Dim det As Double, Ratio As Double
    
    Dx1 = x1 - x2
    Dx2 = x3 - x4
    Dx3 = x4 - x2
    Dy1 = y1 - y2
    Dy2 = y3 - y4
    Dy3 = y4 - y2
    
    det = (Dx2 * Dy1) - (Dy2 * Dx1)
    
    If Abs(det) <= eps Then
        det = (Dx2 * Dy3) - (Dy2 * Dx3)
        If Abs(det) <= eps Then
            IntersectionPoint.X = x3
            IntersectionPoint.Y = y3
            Exit Function
        End If
    End If
    Ratio = ((Dx1 * Dy3) - (Dy1 * Dx3)) / det
    IntersectionPoint.X = (Ratio * Dx2) + x4
    IntersectionPoint.Y = (Ratio * Dy2) + y4
End Function
Public Function Position2(ByVal x1 As Double, ByVal y1 As Double, _
                          ByVal x2 As Double, ByVal y2 As Double, _
                          ByVal Px As Double, ByVal Py As Double) As Position2
    'gibt zurck auf welcher Seite sich ein Punkt P relativ zu einer Geraden G befindet
    'Gerade: G:{{x1, y1}, {x2, y2}}
    'Punkt : P: {Px, Py}
    Dim det As Double
    'Determinante aus den 3 Punkten
    det = (x2 - x1) * (Py - y1) - (Px - x1) * (y2 - y1)
    If Abs(det) <= eps Then
        Position2 = Collinear
    ElseIf det > 0# Then
        Position2 = LeftPosition
    ElseIf det < 0# Then
        Position2 = RightPosition
    End If
End Function
Private Function IsInsideClippingRegion(P As Point2F, s As LineSegment2F) As Boolean
    'bei der Computer-Geometrie taucht immer wieder eine wichtige Frage auf:
    'wo befindet sich ein Punkt relativ zu einer Geraden bzw auf welcher Seite der Geraden
    'liegt ein Punkt. Hier ist ein kartesisches KO-System definiert.
    'Auerdem ist die Umfahrungsrichtung eines Polygons rechts herum definiert.
    
    'RightHandside oder Collinear ist inside
    'ist das KO-System anders definiert so mu hier angepasst werden
    IsInsideClippingRegion = _
        ( _
            MGeo.Position2(s.P(0).X, s.P(0).Y, _
                           s.P(1).X, s.P(1).Y, _
                           P.X, P.Y _
                           ) <> LeftPosition _
        )
End Function
Public Function SutherlandHodgmanClip(Clipper() As Point2F, _
                                      ClippingPoly() As Point2F, _
                                      PolyOut As Polygon2F _
                                      ) As Boolean
    'Clippt ein beliebiges Polygon (konkav oder konvex) an einem konvexen Polygon
    'gibt zurck ob eine Verschneidung geklappt hat
    'Clipper       ein konvexes Polygon das clippt
    'ClippingPoly  das Polygon das geclippt wird
    'PolyOut       das Polygon das zurckgegeben wird
    'konkav: nach innen  gewlbt  ], >, u, v
    'konvex: nach aussen gewlbt  0, o, O,
    
    'Achtung:
    'bevor die Funktion angewendet wird, mu sichergestellt werden
    'da der Datenbereich der Arrays der Polygone Clipper und ClippingPoly
    'nur so gro ist wie die Anzahl Ihrer Punkte ist.
    '
    Dim ClippedPoly As Polygon2F
    Dim CS As LineSegment2F 'ein Segment aus Clipper
    Dim PS As LineSegment2F 'ein Segment aus ClippingPoly
    Dim SP As Point2F   'ein Schnittpunkt aus CS und PS
    Dim i  As Long, i0 As Long, i1 As Long
    Dim j  As Long
    Dim uC As Long
    Dim uP As Long
    uC = UBound(Clipper)
    With PolyOut
        .Count = UBound(ClippingPoly) + 1
        .P = ClippingPoly
    End With
    For j = 0 To uC
        If j < uC Then i0 = j: i1 = j + 1 Else i0 = uC: i1 = 0
        CS.P(0) = Clipper(i0):  CS.P(1) = Clipper(i1)
        'das alte Polygon lschen, alle Punkte werden neu hinzugefgt
        'das hat den Vorteil da die Punkte spter nicht sortiert werden mssen
        ClippedPoly = MGeo.ErasePolygon2F
        uP = PolyOut.Count - 1
        For i = 0 To uP
            If i < uP Then i0 = i: i1 = i + 1 Else i0 = uP: i1 = 0
            PS.P(0) = PolyOut.P(i0):  PS.P(1) = PolyOut.P(i1)
            If IsInsideClippingRegion(PS.P(0), CS) Then
                If IsInsideClippingRegion(PS.P(1), CS) Then
                    'beide drinnen, den zweiten hinzufgen
                    Call MPolygon2F.Add(ClippedPoly, PS.P(1))
                Else
                    'der erste drinnen, der zweite nicht drinnen,
                    'dann den Schnittpunkt berechnen
                    Call MPolygon2F.Add(ClippedPoly, MLineSegment2F.LineIntersection(CS, PS))
                End If
            Else
                'der erste nicht drinnen
                If IsInsideClippingRegion(PS.P(1), CS) Then
                    'der zweite drinnen,
                    'dann den Schnittpunkt berechnen
                    Call MPolygon2F.Add(ClippedPoly, MLineSegment2F.LineIntersection(CS, PS))
                    'und den zweiten hinzufgen
                    Call MPolygon2F.Add(ClippedPoly, PS.P(1))
                End If
            End If
        Next
        'das geclippte Polygon wird der Input fr den nchsten Schritt
        PolyOut = ClippedPoly
    Next
    If PolyOut.Count > 0 Then SutherlandHodgmanClip = True
End Function
